@@ -11,6 +11,35 @@
 #define IS_TRUE(x) (Rf_isLogical(x) && Rf_length(x) && asLogical(x))
 #define IS_FALSE(x) (Rf_isLogical(x) && Rf_length(x) && !asLogical(x))
 
+/* copy from R source */
+
+const char *formatError(DWORD res){
+  static char buf[1000], *p;
+  FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
+                NULL, res,
+                MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
+                buf, 1000, NULL);
+  p = buf+strlen(buf) -1;
+  if(*p == '\n') *p = '\0';
+  p = buf+strlen(buf) -1;
+  if(*p == '\r') *p = '\0';
+  p = buf+strlen(buf) -1;
+  if(*p == '.') *p = '\0';
+  return buf;
+}
+
+
+/* check for system errors */
+void bail_if(int err, const char * what){
+  if(err)
+    Rf_errorcall(R_NilValue, "System failure for: %s (%s)", what, formatError(GetLastError()));
+}
+
+void warn_if(int err, const char * what){
+  if(err)
+    Rf_warningcall(R_NilValue, "System failure for: %s (%s)", what, formatError(GetLastError()));
+}
+
 /* Check for interrupt without long jumping */
 void check_interrupt_fn(void *dummy) {
   R_CheckUserInterrupt();
@@ -39,9 +68,10 @@ static DWORD WINAPI PrintPipe(HANDLE pipe, FILE *stream){
     if(!ReadFile(pipe, buffer, 65337, &len, NULL)){
       int err = GetLastError();
       if(err != ERROR_BROKEN_PIPE)
-        printf("ReadFile(pipe) failed (%d)\n", err);
+        Rprintf("ReadFile(pipe) failed (%d)\n", err);
       CloseHandle(pipe);
       ExitThread(0);
+      return(0);
     }
     fprintf(stream, "%.*s", (int) len, buffer);
   }
@@ -58,8 +88,7 @@ static DWORD WINAPI PrintErr(HANDLE pipe){
 void ReadFromPipe(SEXP fun, HANDLE pipe){
   unsigned long len = 1;
   while(1){
-    if(!PeekNamedPipe(pipe, NULL, 0, NULL, &len, NULL))
-      Rf_errorcall(R_NilValue, "PeekNamedPipe failed");
+    bail_if(!PeekNamedPipe(pipe, NULL, 0, NULL, &len, NULL), "PeekNamedPipe");
     if(!len)
       break;
     char buffer[len];
@@ -75,8 +104,10 @@ HANDLE fd(const char * path){
   sa.lpSecurityDescriptor = NULL;
   sa.bInheritHandle = TRUE;
   DWORD dwFlags = FILE_ATTRIBUTE_NORMAL;
-  return CreateFile(path, GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE,
+  HANDLE out = CreateFile(path, GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE,
                     &sa, CREATE_ALWAYS, dwFlags, NULL);
+  bail_if(out == INVALID_HANDLE_VALUE, "CreateFile");
+  return out;
 }
 
 BOOL CALLBACK closeWindows(HWND hWnd, LPARAM lpid) {
@@ -88,6 +119,21 @@ BOOL CALLBACK closeWindows(HWND hWnd, LPARAM lpid) {
   return TRUE;
 }
 
+void fin_proc(SEXP ptr){
+  if(!R_ExternalPtrAddr(ptr)) return;
+  CloseHandle(R_ExternalPtrAddr(ptr));
+  R_ClearExternalPtr(ptr);
+}
+
+// Keeps one process handle open to let exec_status() read exit code
+SEXP make_handle_ptr(HANDLE proc){
+  SEXP ptr = PROTECT(R_MakeExternalPtr(proc, R_NilValue, R_NilValue));
+  R_RegisterCFinalizerEx(ptr, fin_proc, 1);
+  setAttrib(ptr, R_ClassSymbol, mkString("handle_ptr"));
+  UNPROTECT(1);
+  return ptr;
+}
+
 SEXP C_execute(SEXP command, SEXP args, SEXP outfun, SEXP errfun, SEXP wait){
   int block = asLogical(wait);
   SECURITY_ATTRIBUTES sa;
@@ -103,20 +149,16 @@ SEXP C_execute(SEXP command, SEXP args, SEXP outfun, SEXP errfun, SEXP wait){
 
   //set STDOUT pipe
   if(block || IS_TRUE(outfun)){
-    if (!CreatePipe(&pipe_out, &si.hStdOutput, &sa, 0))
-      Rf_errorcall(R_NilValue, "Failed to creat stdout pipe");
-    if (!SetHandleInformation(pipe_out, HANDLE_FLAG_INHERIT, 0))
-      Rf_errorcall(R_NilValue, "SetHandleInformation failed");
+    bail_if(!CreatePipe(&pipe_out, &si.hStdOutput, &sa, 0), "CreatePipe stdout");
+    bail_if(!SetHandleInformation(pipe_out, HANDLE_FLAG_INHERIT, 0), "SetHandleInformation stdout");
   } else if(IS_STRING(outfun)){
     si.hStdOutput = fd(CHAR(STRING_ELT(outfun, 0)));
   }
 
   //set STDERR
   if(block || IS_TRUE(errfun)){
-    if (!CreatePipe(&pipe_err, &si.hStdError, &sa, 0))
-      Rf_errorcall(R_NilValue, "Failed to creat stdout pipe");
-    if (!SetHandleInformation(pipe_err, HANDLE_FLAG_INHERIT, 0))
-      Rf_errorcall(R_NilValue, "SetHandleInformation failed");
+    bail_if(!CreatePipe(&pipe_err, &si.hStdError, &sa, 0), "CreatePipe stderr");
+    bail_if(!SetHandleInformation(pipe_err, HANDLE_FLAG_INHERIT, 0), "SetHandleInformation stdout");
   } else if(IS_STRING(errfun)){
     si.hStdError = fd(CHAR(STRING_ELT(errfun, 0)));
   }
@@ -130,7 +172,7 @@ SEXP C_execute(SEXP command, SEXP args, SEXP outfun, SEXP errfun, SEXP wait){
   }
   PROCESS_INFORMATION pi = {0};
   if(!CreateProcess(NULL, argv, &sa, &sa, TRUE, CREATE_NO_WINDOW | CREATE_BREAKAWAY_FROM_JOB | CREATE_SUSPENDED, NULL, NULL, &si, &pi))
-    Rf_errorcall(R_NilValue, "Failed to execute '%s'", cmd);
+    Rf_errorcall(R_NilValue, "Failed to execute '%s' (%s)", cmd, formatError(GetLastError()));
 
   //CloseHandle(pi.hThread);
   DWORD pid = GetProcessId(pi.hProcess);
@@ -139,8 +181,7 @@ SEXP C_execute(SEXP command, SEXP args, SEXP outfun, SEXP errfun, SEXP wait){
 
   //A 'job' is some sort of process container
   HANDLE job = CreateJobObject(NULL, NULL);
-  if(!AssignProcessToJobObject(job, proc))
-    Rf_errorcall(R_NilValue, "AssignProcessToJobObject failed: %d", GetLastError());
+  bail_if(!AssignProcessToJobObject(job, proc), "AssignProcessToJobObject");
   ResumeThread(thread);
   CloseHandle(thread);
 
@@ -164,20 +205,44 @@ SEXP C_execute(SEXP command, SEXP args, SEXP outfun, SEXP errfun, SEXP wait){
       }
     }
     DWORD exit_code;
-    CloseHandle(pipe_out);
-    CloseHandle(pipe_err);
-    GetExitCodeProcess(proc, &exit_code);
+    warn_if(!CloseHandle(pipe_out), "CloseHandle pipe_out");
+    warn_if(!CloseHandle(pipe_err), "CloseHandle pipe_err");
+    warn_if(GetExitCodeProcess(proc, &exit_code) == 0, "GetExitCodeProcess");
+    warn_if(!CloseHandle(proc), "CloseHandle proc");
     res = exit_code; //if wait=TRUE, return exit code
   } else {
     //create background threads to print stdout/stderr
     if(IS_TRUE(outfun))
-      CreateThread(NULL, 0, PrintOut, pipe_out, 0, 0);
+      bail_if(!CreateThread(NULL, 0, PrintOut, pipe_out, 0, 0), "CreateThread stdout");
     if(IS_TRUE(errfun))
-      CreateThread(NULL, 0, PrintErr, pipe_err, 0, 0);
+      bail_if(!CreateThread(NULL, 0, PrintErr, pipe_err, 0, 0), "CreateThread stderr");
   }
-  CloseHandle(proc);
   CloseHandle(job);
   CloseHandle(si.hStdError);
   CloseHandle(si.hStdOutput);
-  return ScalarInteger(res);
+  SEXP out = PROTECT(ScalarInteger(res));
+  if(!block)
+    setAttrib(out, install("handle"), make_handle_ptr(proc));
+  UNPROTECT(1);
+  return out;
+}
+
+SEXP R_exec_status(SEXP rpid, SEXP wait){
+  DWORD exit_code = NA_INTEGER;
+  int pid = asInteger(rpid);
+  HANDLE proc = OpenProcess(PROCESS_QUERY_INFORMATION | SYNCHRONIZE, FALSE, pid);
+  bail_if(!proc, "OpenProcess()");
+  do {
+    DWORD res = WaitForSingleObject(proc, 200);
+    bail_if(res == WAIT_FAILED, "WaitForSingleObject()");
+    if(res != WAIT_TIMEOUT)
+      break;
+  } while(asLogical(wait) && !pending_interrupt());
+  warn_if(GetExitCodeProcess(proc, &exit_code) == 0, "GetExitCodeProcess");
+  CloseHandle(proc);
+  return ScalarInteger(exit_code == STILL_ACTIVE ? NA_INTEGER : exit_code);
+}
+
+SEXP R_eval_fork(SEXP call, SEXP env, SEXP subtmp){
+  Rf_error("eval_fork not available on windows");
 }
